package devmgr.api.rpc;

import java.io.IOException;
import java.net.InetAddress;

import devmgr.api.runtime.ApiDefinition;
import devmgr.api.runtime.DecodeException;
import devmgr.api.runtime.EncodeException;
import devmgr.api.runtime.Decoder.TypeOverrideTable;

/**
 * This class implements an RPC client by acting as a wrapper for
 * protocol objects derived from RPCClientGeneric.  It handles both
 * TCP and UDP connections.
 */
public class RPCClient
{
    private ApiDefinition m_ApiDef;
    private RPCClientGeneric m_Client;

    /**
     * Constructor for RPC client objects.
     * @param apiDef API definition for this client.
     */
    public RPCClient(ApiDefinition apiDef)
    {
        m_ApiDef = apiDef;
    }

    /**
     * Connect this client to the specified server.
     * @param server Server's IP address.
     * @param port TCP/UDP port number to connect to.
     * @param stream True for TCP connections, false for UDP.
     * @exception RPCError If any RPC error occurs.
     * @exception IOException If any socket error occurs.
     */
    public void
    connect(InetAddress server, int port, boolean stream)
    throws RPCError, IOException
    {
        // Create the appropriate underlying protocol object, depending on
        // the type of connection requested by the caller.
        if (stream)
            m_Client = new RPCClientTCP(m_ApiDef, server, port);
        else
            m_Client = new RPCClientUDP(m_ApiDef, server, port);
    }

    /**
     * Perform a call/return sequence on the specified remote procedure,
     * using the argument value given by the user.
     * @param proc Procedure being called.
     * @param arg Argument being passed to the procedure.
     * @return Result value from the called procedure.
     * @exception RPCError If any RPC error occurs.
     * @exception IOException If any underlying network error occurs.
     */
    public Object call(int proc, Object arg)
    throws RPCError, IOException, EncodeException, DecodeException
    {
        return m_Client.call(proc, arg);
    }

    /**
     * Close the network connection to the remote server.  No further
     * operations are permitted on this object after it is closed.
     * @exception IOException If any underlying network error occurs.
     */
    public void close()
    throws IOException
    {
        m_Client.close();
        m_Client = null;
        return;
    }

    /**
     * Sets the timeout limit for the underlying connection.  If any
     * receive operation takes longer than the specified time limit,
     * an RPCError exception will be thrown during the call sequence.
     * @param seconds Number of seconds to wait before timing out.
     * @exception IOException If an I/O error occurs on the connection.
     */
    public void setTimeout(int seconds)
    throws IOException
    {
        m_Client.setTimeout(seconds);
        return;
    }

    /**
     * Returns the IP address of the client end of the RPC connection.
     * This is particularly useful for clients running on multi-homed
     * hosts, where it's not obvious what local IP address is being
     * used by the client-side network socket.
     * @return Client-side IP address.
     */
    public InetAddress getLocalAddress()
    {
        return m_Client.getLocalAddress();
    }

    /**
     * Sets the type-override table that is used when decoding reply
     * messages from the remote server.
     * @param table Type override table to use.
     */
    public synchronized void setTypeOverrides(TypeOverrideTable table)
    {
        m_Client.setTypeOverrides(table);
    }

    /**
     * Sets the authentication generator object that will be used by
     * the underlying RPC client when creating outbound RPC messages.
     * This allows the user of the RPC package to provide a customized
     * mechanism for creating authentication data.
     * @param generator New authentication generator object.
     */
    public void setAuthGenerator(RPCAuthGenerator generator)
    {
        m_Client.setAuthGenerator(generator);
        return;
    }

    /**
     * Returns the InetAddress of the connected host.
     * @return Address of the connected host.
     */
    public InetAddress getDestinationIP()
    {
        return m_Client.getDestinationIP();
    }
}
