package devmgr.api.rpc;


import javax.management.MBeanServer;
import javax.management.ObjectName;
import java.io.*;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.lang.management.ManagementFactory;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.concurrent.atomic.AtomicLong;

public class SelectManager implements SelectManagerMBean {
	private Thread checkerThread;
	private AtomicLong maxCheckTime=new AtomicLong();
	private AtomicLong lastCheckTime=new AtomicLong();
	//Map<SelectionKey,RPCClientASyncTCP> readClients=new HashMap<SelectionKey, RPCClientASyncTCP>();
	Selector selector;
    public SelectManager() throws IOException{
        this("default");
    }
	public SelectManager(String id) throws IOException {
		super();
		selector=Selector.open();
		boolean b=selector.isOpen();
		Runnable run=new Runnable(){
			public void run(){
				monitor();
			}
		};
		if (b){
			checkerThread=new Thread(run,"Storage Device NIO Selector Thread: " + id);
			checkerThread.setDaemon(true);
			checkerThread.start();
		}
		
	}

    @Override
	public int getNumberOfChannels(){
		int ret;
		synchronized (selector) {
			ret=selector.keys().size();
		}
		return ret;
	}
	public void registerClient(NIORPCClient client) throws ClosedChannelException{
		SocketChannel channel = client.getChannel();
		try{
			synchronized (selector) {
				SelectionKey key = channel.register(selector, SelectionKey.OP_READ | SelectionKey.OP_CONNECT);
				key.attach(client);
			}
		}finally{
		}
	}
	public void deregisterClient(NIORPCClient client){
		synchronized (selector) {
			Set<SelectionKey> keys = selector.keys();
			for (SelectionKey x:keys){
				Object ob=x.attachment();
				if (ob != null){
                    NIORPCClient tmp=(NIORPCClient) ob;
					if (tmp.equals(client)){
						x.cancel();
						break;
					}
				}
			}
		}
	}
	
	private void monitor(){
		while (!Thread.currentThread().isInterrupted()){
			try {
				int count = 0;
				synchronized (selector) {
					count = selector.select(100);
				}
				checkChannels(count==0);
				Thread.sleep(50);
			}
            catch (InterruptedException ee){
                Thread.currentThread().interrupt();
                //let it die
            }
            catch (Throwable e) {
                //TODO, Somthing better than this
				e.printStackTrace();
            }
		}
	}
	private void checkChannels(boolean onlyCheckTimeout) {
		long t1=System.currentTimeMillis();
		
		List<SelectionKey> tmpkeys=new ArrayList<SelectionKey>();
		if (!onlyCheckTimeout){
			synchronized (selector) {
				Iterator<SelectionKey> it ;
				Set<SelectionKey> selectedKeys = selector.selectedKeys();
				it = selectedKeys.iterator();
				while (it.hasNext()){
					tmpkeys.add(it.next());
				}
				selectedKeys.clear();
			}
			
			for (SelectionKey key:tmpkeys){
                NIORPCClient client = (NIORPCClient) key.attachment();
				try{
					if (key.isConnectable()){
						client.handleConnectSocketSignal();
					}else if (key.isReadable()){
						client.handleReadSignal();
					}
				}catch (Exception t){
                    client.handleError(t);
				}
			}
		}
		
		tmpkeys=new ArrayList<SelectionKey>();
		synchronized (selector) {
			Set<SelectionKey> keys = selector.keys();
			Iterator<SelectionKey> it  = keys.iterator();
			while (it.hasNext()){
				tmpkeys.add(it.next());
			}
		}
		
		long currentTime=System.currentTimeMillis();
		for(SelectionKey key:tmpkeys){
			NIORPCClient client = (NIORPCClient) key.attachment();
			//if we are activly running a command
			if ( client.isActiveCall()){
				long timeout=client.getTimeout();
				timeout*=1000; //turn it into ms
				long lastTouched=client.getLastTouched();
                long diff = currentTime - lastTouched;
                if ( diff > timeout){
					client.handleTimeout();
				}
			}
		}
		
		long t2=System.currentTimeMillis();
		long t3=t2-t1;
		long max=this.maxCheckTime.get();
		if (t3 >max){
			maxCheckTime.set(t3);
		}
		lastCheckTime.set(t3);
	}
	

    public void stop(){
        checkerThread.interrupt();
        //checkerThread.stop();
        synchronized (selector) {
            Set<SelectionKey> keys = selector.keys();
            Iterator<SelectionKey> it  = keys.iterator();
            while (it.hasNext()){
                SelectionKey next = it.next();
                next.cancel();
            }
        }
    }
	/**
	 * Get max time in MS of check process
	 * @return
	 */
    @Override
	public long getMaxCheckTime(){
		return maxCheckTime.get();
	}
    @Override
	public long getLastCheckTime(){
		return lastCheckTime.get();
	}

    @Override
    public void dumpCreationStack(String path) throws FileNotFoundException, UnsupportedEncodingException {

		List<Throwable> traces = new ArrayList<>();
		synchronized (selector) {
			Set<SelectionKey> keys = selector.keys();
			for (SelectionKey x : keys) {
				Object attachment = x.attachment();
				if (attachment != null) {
					NIORPCClient client = (NIORPCClient) attachment;
					traces.add(client.getCreationStack());
				}
			}
		}
		PrintStream out = new PrintStream(new FileOutputStream(path), true, StandardCharsets.UTF_8.name());
		for (Throwable t : traces) {
			t.printStackTrace(out);
		}

		out.flush();
		out.close();
	}
}
