package devmgr.api.rpc;

import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.Socket;

import devmgr.api.runtime.XdrMessage;

/**
 * This class implements a TCP channel for carrying RPC messages.
 * Its methods are responsible for adding/removing RPC 4-byte
 * message length codes, etc., so as to provide the user with
 * a more abstract channel that is used to send/receive basic
 * messages without worrying about the TCP-oriented details.
 */
public class TCPChannel
{
    private Socket m_Sock = null;
    private OutputStream m_Out = null;
    private InputStream m_In = null;

    /**
     * Constructor for the RPC channel class for TCP connections.
     * Just prepares local resources for future use.
     * @param sock Socket to be used for the channel.
     * @exception IOException If any I/O error occurs.
     */
    public TCPChannel(Socket sock)
            throws IOException
    {
        m_Sock = sock;
        m_Sock.setTcpNoDelay(true);

        m_Out = m_Sock.getOutputStream();
        m_In = m_Sock.getInputStream();

        return;
    }

    /**
     * Finalizes objects just prior to garbage collection.
     * @exception Throwable If any underlying exception occurs.
     */
    @Override
    protected void finalize()
            throws Throwable
    {
        try {
            if (m_In != null) {
                m_In.close();
            }

            if (m_Out != null) {
                m_Out.close();
            }

            if (m_Sock != null) {
                m_Sock.close();
            }
        }
        finally {
            super.finalize();
        }
    }

    /**
     * Sets the timeout limit for the underlying connection.
     * @param seconds Number of seconds to wait before timing out.
     * @exception IOException If any I/O error occurs.
     */
    public void setTimeout(int seconds)
            throws IOException
    {
        m_Sock.setSoTimeout(seconds * 1000);
        return;
    }

    /**
     * Closes the underlying network connection.
     * @exception IOException If any I/O error occurs.
     */
    public void close()
            throws IOException
    {
        IOException exception = null;
        try {
            m_In.close();
            m_In = null;
        }catch (IOException e){
            exception = e;
        }

        try{
            m_Out.close();
            m_Out = null;
        }catch (IOException e){
            exception = e;
        }

        try{
            m_Sock.close();
            m_Sock = null;
        }catch (IOException e){
            exception = e;
        }

        if(exception != null) {
            throw exception;
        }

        return;
    }

    /**
     * Returns the IP address of the client end of the RPC connection.
     * @return Client-side IP address.
     */
    public InetAddress getLocalAddress()
    {
        return m_Sock.getLocalAddress();
    }

    /**
     * Sends an outbound RPC message.
     * @param msg The message to be sent.
     * @exception RPCError If an RPC error occurs.
     * @exception IOException If any I/O error occurs.
     */
    public void sendMsg(XdrMessage msg)
            throws RPCError, IOException
    {
        int totalSize = msg.getMessageSize();
        int amtDone = 0;
        int nextBuf = 0;
        byte[] header = new byte[4];

        while (amtDone < totalSize)
        {
            byte[] buf = msg.getBuffer(nextBuf++);
            int amt = Math.min(buf.length, totalSize - amtDone);

            int lengthCode = amt;
            if ((amtDone + amt) >= totalSize)
            {
                lengthCode |= 0x80000000;
            }

            // Set up the TCP record header for this chunk.
            header[0] = (byte) (lengthCode >>> 24);
            header[1] = (byte) (lengthCode >>> 16);
            header[2] = (byte) (lengthCode >>>  8);
            header[3] = (byte) (lengthCode >>>  0);

            // Write the header and data to the output stream.
            m_Out.write(header);
            m_Out.write(buf, 0, amt);

            amtDone += amt;
        }

        // Force transmission of any buffered data.
        m_Out.flush();

        return;
    }

    /**
     * Receives an inbound RPC message and returns it to the caller.
     * @return The received message
     * @exception RPCError If an RPC error occurs.
     * @exception IOException If any I/O error occurs.
     */
    public XdrMessage receiveMsg()
            throws RPCError, IOException
    {
        XdrMessage msg = new XdrMessage();
        byte[] header = new byte[4];
        boolean done = false;
        int amt;

        while (! done)
        {
            readFully(header, 0, header.length);
            int lengthCode = (((int)(header[0]) & 0xFF) << 24) |
                    (((int)(header[1]) & 0xFF) << 16) |
                    (((int)(header[2]) & 0xFF) <<  8) |
                    (((int)(header[3]) & 0xFF) <<  0) ;

            if ( (amt = (lengthCode & 0x7FFFFFFF)) != 0 )
            {
                byte[] buf = new byte[amt];
                readFully(buf, 0, amt);
                msg.addBuffer(buf);
            }

            done = ( (lengthCode & 0x80000000) != 0 );
        }

        return msg;
    }

    /**
     * Receives the specified amount of data from the connection,
     * possibly iterating and accumulating partial results.
     * @param buf Buffer into which data is placed.
     * @param off Buffer offset at which to start filling.
     * @param amt Amount of data to receive.
     * @exception RPCError If an RPC error occurs.
     * @exception IOException If any I/O error occurs.
     */
    private void readFully(byte[] buf, int off, int amt)
            throws RPCError, IOException
    {
        int done = 0;
        int n;

        while (done < amt)
        {
            //Use a try/catch block so we can correctly report RPC TIMEOUTs.
            try
            {
                n = m_In.read(buf, off + done, amt - done);
            }
            catch (InterruptedIOException ex)
            {
                throw new RPCError(RPCError.ErrorType.TIMEOUT);
            }

            if (n <= 0)
            {
                throw new RPCError(RPCError.ErrorType.CONNECTION_LOST);
            }

            done += n;
        }

        return;
    }

    /**
     * Get the InetAddress of the remote (server) system to which this
     * channel is connected.
     * @return Remote server address.
     */
    public InetAddress getServerAddress()
    {
        return m_Sock.getInetAddress();
    }
}
