package devmgr.api.rpc;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;

import devmgr.api.runtime.XdrMessage;

/**
 * This class implements a UDP channel for carrying RPC messages.
 * It provides core methods that can be used by both client and
 * server classes for sending and receiving RPC messages via UDP.
 */
public class UDPChannel
{
    private static final int MAX_MSG_SIZE = 8800;
    private DatagramSocket m_Sock = null;

    /**
     * Constructor for the RPC channel class for UDP connections.
     * Just prepares local resources for future use.
     * @param sock Socket to be used for data transfer.
     */
    public UDPChannel(DatagramSocket sock)
    {
        m_Sock = sock;
    }

    /**
     * Finalizes objects just prior to garbage collection.
     * @exception Throwable If any underlying exception occurs.
     */
    @Override
    protected void finalize()
    throws Throwable
    {
        try {
            if (m_Sock != null) {
                m_Sock.close();
            }
        }
        finally {
            super.finalize();
        }
    }

    /**
     * Sets the timeout limit for the underlying connection.
     * @param millisec Number of milliseconds to wait before timing out.
     * @exception IOException If an underlying I/O error occurs.
     */
    public void setTimeout(int millisec)
    throws IOException
    {
        m_Sock.setSoTimeout(millisec);
    }

    /**
     * Closes the underlying network connection/socket.
     * @exception IOException If an underlying I/O error occurs.
     */
    public void close()
    throws IOException
    {
        m_Sock.close();
        m_Sock = null;
    }
    
    /**
     * Returns the IP address of the client end of the RPC connection.
     * @return Client-side IP address.
     */
    public InetAddress getLocalAddress()
    {
        return m_Sock.getLocalAddress();
    }

    /**
     * Sends an outbound message, which is provided in the form
     * of a UDPMessageBundle object by the caller.
     * @param bundle The message to be sent.
     * @exception RPCError If any RPC error occurs.
     * @exception IOException If an underlying I/O error occurs.
     */
    public void sendMsg(UDPMessageBundle bundle)
    throws RPCError, IOException
    {
        XdrMessage msg = bundle.getMsg();
        int totalSize = msg.getMessageSize();
        DatagramPacket pkt = null;
        byte[] buf;

        // Get one contiguous buffer with all message data inside it.
        buf = msg.getContiguousBuffer();

        // Create a new datagram packet and send it out.
        pkt = new DatagramPacket(buf, totalSize,
            bundle.getAddress(),
            bundle.getPort());
        m_Sock.send(pkt);
    }

    /**
     * Receives an inbound message, returning it in the form
     * of a UDPMessageBundle object.
     * @return A message bundle object containing the inbound message data.
     * @exception RPCError If any RPC error occurs.
     * @exception IOException If an underlying I/O error occurs.
     */
    public UDPMessageBundle receiveMsg()
    throws RPCError, IOException
    {
        XdrMessage msg = new XdrMessage();
        byte[] buf = new byte[MAX_MSG_SIZE];
        DatagramPacket pkt = new DatagramPacket(buf, MAX_MSG_SIZE);
        UDPMessageBundle bundle;

        // Wait for a message to arrive.
        m_Sock.receive(pkt);

        // Add the lone buffer as the last in the IOV.
        msg.addBuffer(buf, pkt.getLength());

        // Create a bundle object and return it to the caller.
        bundle = new UDPMessageBundle(msg, pkt.getAddress(), pkt.getPort());

        return bundle;
    }
}
