package devmgr.api.runtime;

import devmgr.api.symbol.ReturnCode;

import java.util.EnumSet;
import java.util.HashMap;
import java.util.Set;

public class ApiDefinition
{
    private int m_ProgNumber;
    private int m_ProgVersion;
    
    private TypeSpec[] m_TypeTable = null;
    private ProcTableEntry[] m_ProcTable = null;
    
    private HashMap<String, ProcTableEntry> m_ProcsByName = null;
    private HashMap<Integer, ProcTableEntry> m_ProcsByNumber = null;
    
    // We use the following map so that we can share the maps defined
    // above across multiple instances of a given API definition.  This
    // is obviously better than creating many duplicates of maps that
    // can get pretty large.
    private static HashMap<TypeSpec[], ApiDefinition> m_Tables =
        new HashMap<TypeSpec[], ApiDefinition>();
    
    /**
     * This class defines the content of a procedure table entry, which
     * contains all relevant information about a particular callable
     * procedure.
     */
    public static class ProcTableEntry
    {
        public String procName;
        public int procNumber;
        
        public int suggestedTimeout;
        public boolean requiresAuthentication;
        
        public TypeSpec argType;
        public TypeSpec resultType;
        public String description;
        public String argDescription;
        public String resultDescription;
        public String includeFor;
        String obsoletedIn;
        String deprecatedFor;
        public Set<ReturnCode> documentedReturnCodes;
        public ProcTableEntry(String name, int number,
                              int timeout, boolean reqAuth,
                              TypeSpec arg, TypeSpec result, String desc, String argDesc, String resultDesc,
                              String includedFor, String obsoletedIn, String deprecatedFor,
                              final Set<ReturnCode> documentedReturnCodes)
        {
            this.procName = name;
            this.procNumber = number;
            this.suggestedTimeout = timeout;
            this.requiresAuthentication = reqAuth;
            this.argType = arg;
            this.resultType = result;
            this.description=desc;
            this.argDescription = argDesc;
            this.resultDescription = resultDesc;
            this.includeFor=includedFor;
            this.obsoletedIn=obsoletedIn;
            this.deprecatedFor=deprecatedFor;
            this.documentedReturnCodes = documentedReturnCodes;
        }
    }
    
    /**
     * Constructor.
     * @param progNum The defined program's number.
     * @param progVers The defined program's version.
     * @param typeTable The type table for the defined program.
     * @param procTable The procedure table for the defined program.
     */
    public 
    ApiDefinition(int progNum, int progVers, 
                  TypeSpec[] typeTable, ProcTableEntry[] procTable)
    {
        m_ProgNumber = progNum;
        m_ProgVersion = progVers;
        
        m_TypeTable = typeTable;
        m_ProcTable = procTable;
        
        // Protect against concurrent access that might cause 
        // competing threads to simultaneously initialize the maps.
        synchronized(m_Tables)
        {
            // If we already instantiated maps for the specified table,
            // reuse them for this instance.
            ApiDefinition instance = m_Tables.get(m_TypeTable);
            if (instance != null)
            {
                m_ProcsByName = instance.m_ProcsByName;
                m_ProcsByNumber = instance.m_ProcsByNumber;
            }
            
            // Otherwise, build up the maps for the first time here.
            else
            {
                m_ProcsByName = new HashMap<String, ProcTableEntry>();
                m_ProcsByNumber = new HashMap<Integer, ProcTableEntry>();

                if (m_ProcTable != null)
                {
                    for (ProcTableEntry p : m_ProcTable)
                    {
                        m_ProcsByName.put(p.procName, p);
                        m_ProcsByNumber.put(p.procNumber, p);
                    }
                }
            
                // Track for future sharing.
                m_Tables.put(m_TypeTable, this);
            }
        }
    }
    
    /**
     * Getter for the program number.
     * @return Program number.
     */
    public int
    getProgramNumber()
    {
        return m_ProgNumber;
    }
    
    /**
     * Getter for the program version.
     * @return Program version.
     */
    public int
    getProgramVersion()
    {
        return m_ProgVersion;
    }
    
    /**
     * Return the ProcTableEntry for the procedure with the given name.
     * @param name Name of the desired procedure.
     * @return Corresponding ProcTableEntry.
     */
    public ProcTableEntry
    getProcByName(String name)
    {
        return m_ProcsByName.get(name);
    }
    
    /**
     * Return the ProcTableEntry for the procedure with the given number.
     * @param n Number of the desired procedure.
     * @return Corresponding ProcTableEntry.
     */
    public ProcTableEntry
    getProcByNumber(int n)
    {
        return m_ProcsByNumber.get(n);
    }
    
    /**
     * Return the TypeSpec element that describes the specified class.
     * @param cls Class of object being used.
     * @return Corresponding TypeSpec entry for that object.
     */
    public TypeSpec
    getTypeSpec(Class<?> cls)
    {
        TypeTableIndex tti = cls.getAnnotation(TypeTableIndex.class);
        if (tti == null)
            return null;
        
        return m_TypeTable[tti.value()];
    }

    public HashMap<Integer, ProcTableEntry> getMapOfProcNumbersToEntries(){
        return m_ProcsByNumber;
    }
}