package devmgr.api.runtime;

import java.lang.annotation.Annotation;
import java.lang.reflect.Array;
import java.lang.reflect.Field;

/**
 * This class implements an encoder for objects that are defined
 * with RPC-oriented data types (e.g. opaque byte arrays, discriminated
 * unions, fixed/variable-length arrays, etc.).  It is abstract, as it
 * requires a subclass that actually generates encoded objects for a
 * particular representation type (e.g. JSON, XML, etc.).  The core logic
 * for driving the encoding operations resides here.
 */
public abstract class Encoder
{
    public static enum Context
    {
        NONE,
        ARRAY_FIRST,
        ARRAY_NONFIRST,
        STRUCT_FIRST,
        STRUCT_NONFIRST,
        UNION_ALTERNATE
    }

    private static Class<? extends Annotation> m_HiddenFieldAnnotation =
        javax.xml.bind.annotation.XmlTransient.class;
    
    public String m_DefaultRootName = "root";
    private boolean m_EncodeNulls = false;
    private boolean m_EncodeHiddenFields = false;
    
    // These are all of the abstract methods that have to be
    // implemented by subclasses that are focused on encoding
    // to a particular representation.

    protected abstract void
    flush()
    throws EncodeException;
    
    protected abstract void 
    encodeByte(byte obj, String name, Context ctx)
    throws EncodeException;
    
    protected abstract void 
    encodeShort(short obj, String name, Context ctx)
    throws EncodeException;
    
    protected abstract void 
    encodeInt(int obj, String name, Context ctx)
    throws EncodeException;
    
    protected abstract void 
    encodeLong(long obj, String name, Context ctx)
    throws EncodeException;
    
    protected abstract void 
    encodeBoolean(boolean obj, String name, Context ctx)
    throws EncodeException;
    
    protected abstract void 
    encodeFloat(float obj, String name, Context ctx)
    throws EncodeException;
    
    protected abstract void 
    encodeDouble(double obj, String name, Context ctx)
    throws EncodeException;
    
    protected abstract void 
    encodeString(String obj, String name, Context ctx,
                 int maxLength)
    throws EncodeException;
    
    protected abstract void 
    encodeOpaque(String data, String name, Context ctx, OpaqueSpec typeSpec)
    throws EncodeException;
    
    protected abstract void 
    encodeArrayStart(String name, Context ctx,
                     boolean isVariable, 
                     int actualLength)
    throws EncodeException;
    
    protected abstract void 
    encodeArrayEnd(String name, Context ctx,
                   boolean isVariable, 
                   int actualLength)
    throws EncodeException;
    
    protected abstract void 
    encodeStructStart(String name, Context ctx, StructSpec spec)
    throws EncodeException;
    
    protected abstract void 
    encodeStructEnd(String name, Context ctx, StructSpec spec)
    throws EncodeException;
    
    protected abstract void 
    encodeUnionStart(String name, 
                     Context ctx,
                     UnionSpec spec,
                     EnumValue discrimValue)
    throws EncodeException;
    
    protected abstract void 
    encodeUnionEnd(String name, Context ctx, UnionSpec spec)
    throws EncodeException;
    
    protected abstract void 
    encodeEnum(EnumValue value, String name, Context ctx)
    throws EncodeException;
    
    protected abstract void 
    encodePointerPrefix(String name, Context ctx, boolean isNull)
    throws EncodeException;

    /**
     * This method provides a default encoding implementation for
     * encoders that implement enumeration sets as standard arrays
     * of the subject enumeration.  They can use it instead of
     * providing their own implementation.  Otherwise, the subclass
     * must override with a different implementation.
     * @param values
     * @param name
     * @param ctx
     * @throws EncodeException
     */
    protected void
    encodeEnumSet(EnumValue[] values, String name, Context ctx)
    throws EncodeException
    {
        encodeArrayStart(name, ctx, true, values.length);

        Context innerCtx = Context.ARRAY_FIRST;
        for (int i = 0; i < values.length; i++)
        {
            encodeEnum(values[i], name, innerCtx);
            innerCtx = Context.ARRAY_NONFIRST;
        }
        
        encodeArrayEnd(name, ctx, true, values.length);
    }    
    
    public void
    setEncodeNulls(boolean encodeNulls)
    {
        m_EncodeNulls = encodeNulls;
    }
    
    public void
    setEncodeHiddenFields(boolean encodeHiddenFields)
    {
        m_EncodeHiddenFields = encodeHiddenFields;
    }
    
    public void
    encode(Object obj, TypeSpec type)
    throws EncodeException
    {
        processByType(obj, m_DefaultRootName, Context.NONE, type);
        flush();
    }
    
    public void
    encode(Object obj, ApiDefinition api)
    throws EncodeException
    {
        processByType(obj, m_DefaultRootName, Context.NONE, 
                      api.getTypeSpec(obj.getClass()));
        flush();
    }
    
    /**
     * Process an object for which the name and type are both
     * already known.
     * @param obj The entity to be processed.
     * @param name The name of the entity.
     * @param ctx The encoding context.
     * @param objType The entity's type information.
     */
    protected void
    processByType(Object obj, String name, 
                  Context ctx, TypeSpec objType)
    throws EncodeException
    {
        // A null type means this is a void entity, e.g. for an
        // empty argument list in a procedure call.  So we can
        // just bail immediately.
        if (objType == null)
            return;
        
        // Ignore null objects for all EXCEPT pointers.
        if ( (! m_EncodeNulls) &&
             (obj == null) && 
             (objType.typeCode != TypeSpec.TypeCode.POINTER) ) 
            return;

        // Process based on object type.
        switch (objType.typeCode)
        {
            case BYTE:
                encodeByte((byte) obj, name, ctx);
                break;
                
            case CHAR:
                int x = (int) ((Character) obj).charValue();
                encodeByte((byte) (x & 0xFF), name, ctx);
                break;
                
            case SHORT:
                encodeShort((short) obj, name, ctx);
                break;
             
            case INT:
                encodeInt((int) obj, name, ctx);
                break;
                
            case LONG:
                encodeLong((long) obj, name, ctx);
                break;
                
            case BOOLEAN:
                encodeBoolean((boolean) obj, name, ctx);
                break;
                
            case FLOAT:
                encodeFloat((float) obj, name, ctx);
                break;
                
            case DOUBLE:
                encodeDouble((double) obj, name, ctx);
                break;
                
            case STRING:
                processString((String) obj, name, 
                              ctx, (StringSpec) objType);
                break;
                
            case OPAQUE:
                processOpaque((String) obj, name, 
                              ctx, (OpaqueSpec) objType);
                break;
                
            case ARRAY:
                processArray(obj, name, ctx, (ArraySpec) objType);
                break;
                
            case ENUM:
                processEnum(obj, name, ctx, (EnumSpec) objType);
                break;
                
            case POINTER:
                encodePointerPrefix(name, ctx, obj == null);
                processByType(obj, name, ctx, 
                              ((PointerSpec) objType).targetType);
                break;
                
            case STRUCT:
                processStruct(obj, name, ctx, (StructSpec) objType);
                break;
                
            case UNION:
                processUnion(obj, name, ctx, (UnionSpec) objType);
                break;
            
            case SUBCLASS:
                processSubclass(obj, name, ctx, (SubclassSpec) objType);
                break;
            
            case TYPEDEF:
                processTypedef(obj, name, ctx, (TypedefSpec) objType);
                break;
                
            case ENUMSET:
                processEnumSet(obj, name, ctx, (EnumSetSpec) objType);
        }
    }

    /**
     * This method processes a string.
     * @param obj String to be processed.
     * @param name Name of the string field.
     * @param ctx Encoding context.
     * @param typeSpec Type specification for the string.
     * @throws EncodeException
     */
    private void 
    processString(String obj, String name, Context ctx, 
                  StringSpec typeSpec)
    throws EncodeException
    {
        encodeString(obj, name, ctx, typeSpec.maxLength);
    }
    
    /**
     * This method processes an opaque byte array.
     * @param obj Opaque byte array to be processed.
     * @param name Name of the object.
     * @param ctx Encoding context.
     * @param typeSpec Type specification for the object.
     * @throws EncodeException
     */
    private void 
    processOpaque(String obj, String name, Context ctx, OpaqueSpec typeSpec)
    throws EncodeException
    {
        encodeOpaque(obj, name, ctx, typeSpec);
    }
    
    /**
     * This method processes an enumeration set.
     * @param obj Array of enumeration values.
     * @param name Name of the object.
     * @param ctx Encoding context.
     * @param enumSpec Type specification for the enumeration set.
     * @throws EncodeException
     */
    private void
    processEnumSet(Object obj, String name, Context ctx, 
                   EnumSetSpec typeSpec)
    throws EncodeException
    {
        int count = (obj == null) ? 0 : Array.getLength(obj);
        
        // Convert the true enumeration objects into our intermediate
        // format so we can feed to the subclass code that processes it.
        EnumValue[] values = new EnumValue[count]; 
        for (int i = 0; i < values.length; i++)
            values[i] = getEnumValue(Array.get(obj, i), typeSpec.enumType);
        
        encodeEnumSet(values, name, ctx);
    }

    /**
     * This method processes an enumeration.
     * @param obj Enumeration to be processed.
     * @param name Name of the object.
     * @param ctx Encoding context.
     * @param enumSpec Type specification for the object.
     * @throws EncodeException
     */
    private void
    processEnum(Object obj, String name, Context ctx, EnumSpec typeSpec)
    throws EncodeException
    {
        EnumValue evi;
        if (obj != null)
            evi = getEnumValue(obj, typeSpec);
        else
            evi = new EnumValue(ApiUtil.BOGUS_ENUM_NAME, 
                                ApiUtil.BOGUS_ENUM_VALUE);
        
        encodeEnum(evi, name, ctx);
    }

    /**
     * This method processes an array.  It emits up-calls to the
     * methods that bracket the start and end of an array.  In
     * between those two, it emits per-element calls to encode
     * the array items themselves.  Those per-element calls are
     * made with the name argument set to null, as that is how
     * we indicate that the encoding is to take place in the
     * array context.
     * @param array Array to be processed.
     * @param name Name of the array itself.
     * @param ctx Encoding context.
     * @param objType Type specification for the array.
     */
    private void
    processArray(Object array, String name, Context ctx, 
                 ArraySpec objType)
    throws EncodeException
    {
        int actualLength = (array == null) ? 0 : Array.getLength(array);

        // Verify length of fixed-length arrays.
        if ( (! objType.isVariable) &&
             (actualLength != objType.maxLength) )
        {
            String clsName = array != null ? array.getClass().getCanonicalName() : "null";
            throw EncodeException.build("%s: length %d, should be %d",
                    clsName, actualLength, objType.maxLength);
        }

        // Verify limit of variable-length arrays.
        if ( objType.isVariable &&
             (objType.maxLength != 0) &&
             (actualLength > objType.maxLength) )
        {
            String clsName = array != null ? array.getClass().getCanonicalName() : "null";
            throw EncodeException.build("%s: length %d > max %d",
                    clsName, actualLength, objType.maxLength);
        }
        
        encodeArrayStart(name, ctx, objType.isVariable, actualLength);

        // Process each element in the array.
        Context itemContext = Context.ARRAY_FIRST;
        for (int i = 0; i < actualLength; i++)
        {
            try
            {
                processByType(Array.get(array, i), name, 
                              itemContext, objType.elementType);
                itemContext = Context.ARRAY_NONFIRST;
            }
            catch (ArrayIndexOutOfBoundsException ex)
            {
            }
            catch (IllegalArgumentException ex)
            {
            }
        }
        
        encodeArrayEnd(name, ctx, objType.isVariable, actualLength);
    }

    /**
     * This method encodes a structure.  It emits up-calls to the
     * methods that bracket the start and end of the structure.
     * In between those two, it emits per-field calls to encode
     * the structure fields themselves.
     * @param obj Structure to be processed.
     * @param name Name of the structure itself.
     * @param ctx Encoding context.
     * @param objType Type specification for the structure.
     */
    private void
    processStruct(Object obj, String name, Context ctx, StructSpec objType)
    throws EncodeException
    {
        encodeStructStart(name, ctx, objType);
        
        // If this is a "collapsed" structure, we just want to make
        // it look like the structure wrappings are not there.  The
        // "struct" is really more like an alias for its one/only
        // defined field.
        if (objType.isCollapsed)
        {
            processByType(obj, name, ctx, objType.fields[0].typeSpec);
        }
        
        // For non-collapsed structures, process field-by-field.
        else if (obj != null)
        {
            Context itemContext = Context.STRUCT_FIRST;
            for (int i = 0; i < objType.fields.length; i++)
            {
                TypeBinding binding = objType.fields[i];
                
                try
                {
                    Field fld = 
                        objType.cls.getDeclaredField(binding.itemName);
                
                    if ( (! m_EncodeHiddenFields) &&
                         fld.isAnnotationPresent(m_HiddenFieldAnnotation) )
                        continue;
                    
                    fld.setAccessible(true);
                    processByType(fld.get(obj), binding.itemName, 
                                  itemContext, binding.typeSpec);
                    itemContext = Context.STRUCT_NONFIRST;
                }
                catch (Exception caught)
                {
                    throw buildException(caught, obj);
                }
            }
        }
        
        encodeStructEnd(name, ctx, objType);
    }
            
    /**
     * This method encodes a union.  It emits up-calls to the
     * methods that bracket the start and end of the union, then
     * handles encoding of the discriminator and the data value.
     * @param obj Union to be processed.
     * @param name Name of the union itself.
     * @param ctx Encoding context.
     * @param objType Type specification for the union.
     */
    private void
    processUnion(Object obj, String name, Context ctx, UnionSpec objType)
    throws EncodeException
    {
        Class<?> cls = objType.cls;
        
        try
        {
            // Dig out the discriminator enumeration's type specifier. 
            EnumSpec enumType = (EnumSpec) objType.discrimType;
            
            // Extract the case value from the discriminator field.
            EnumValue discrimVal;
            if (obj == null)
                discrimVal = new EnumValue(ApiUtil.BOGUS_ENUM_NAME,
                                           ApiUtil.BOGUS_ENUM_VALUE);
            else
            {
                Field discrimField = 
                    cls.getDeclaredField(objType.discrimName);
                
                discrimField.setAccessible(true);
                discrimVal = getEnumValue(discrimField.get(obj), enumType);
            }
             
            // Start encoding the union, which presumably will make use
            // of the discriminator name and value.
            encodeUnionStart(name, ctx, objType, discrimVal);
            
            // Find the relevant alternate for this case.
            TypeBinding altBinding = objType.caseMap.get(discrimVal.value);
            
            // Use the default binding if necessary.
            if (altBinding == null)
                altBinding = objType.defaultBinding;
            
            // Encode only if we found a binding, and it is non-void; 
            // otherwise assume "void".
            if ( (altBinding != null) && (altBinding.itemName != null) )
            {
                Field altField = cls.getDeclaredField(altBinding.itemName);
                altField.setAccessible(true);
                Object altValue = altField.get(obj);
                
                processByType(altValue,
                              altBinding.itemName,
                              Context.UNION_ALTERNATE,
                              altBinding.typeSpec);
            }
            
            // Terminate the encoding process for the union.
            encodeUnionEnd(name, ctx, objType); 
        }
        
        catch (Exception caught)
        {
            throw buildException(caught, obj);
        }
    }
    
    /**
     * This method processes a subclass.  Since subclasses are always
     * just aliases for a base class, we simply re-process the request
     * using the base class type specification.
     * @param obj Object to be processed.
     * @param name Name of the object.
     * @param ctx Encoding context.
     * @param typeSpec Type specification for the object.
     * @throws EncodeException
     */
    private void
    processSubclass(Object obj, String name, Context ctx, SubclassSpec objType)
    throws EncodeException
    {
        processByType(obj, name, ctx, objType.superType);
    }
    
    /**
     * This method processes a typedef, which contains a single field
     * named "value" whose type is tracked in the type specification.
     * @param obj Object to be processed.
     * @param name Name of the object.
     * @param ctx Encoding context.
     * @param typeSpec Type specification for the object.
     * @throws EncodeException
     */
    private void
    processTypedef(Object obj, String name, Context ctx, TypedefSpec objType)
    throws EncodeException
    {
        try
        {
            Object targetObj = null;
            
            if (obj != null)
            {
                Field fld = objType.cls.getDeclaredField("value");
                targetObj = fld.get(obj);
            }
            
            processByType(targetObj, name, ctx, objType.typeSpec);
        }
        
        catch (Exception caught)
        {
            throw buildException(caught, obj);
        }
    }
    
    /**
     * This method extracts the value of an enumeration object and
     * returns it to the caller.
     * @param obj The enumeration object.
     * @param typeSpec The type specifier for the enumeration object.
     * @return An EnumValue object with the value and enum label.
     */
    private EnumValue
    getEnumValue(Object obj, EnumSpec typeSpec)
    throws EncodeException
    {
        int val;
        String str = null;
        
        try
        {
            Field valueField = typeSpec.cls.getDeclaredField("value");
            valueField.setAccessible(true);

            val = (int) valueField.get(obj);
            str = typeSpec.valueToNameMap.get(val);
        }
        catch (Exception caught)
        {
            throw buildException(caught, obj);
        }
        
        if (str == null)
        {
            throw EncodeException.build("Enum type %s, bad value '%d'",
                                        typeSpec.cls.getCanonicalName(),
                                        val);
        }
        
        return new EnumValue(str, val);
    }

    /**
     * This method builds an exception to be thrown back to the original
     * caller, including information about the class being processed.
     * @param ex Original exception that was caught.
     * @param cls Object type being processed.
     * @return An DecodeException.
     */
    private EncodeException
    buildException(Exception ex, Object obj)
    {
        String name = obj != null ? obj.getClass().getCanonicalName() : "null";
        return EncodeException.build(ex, "Processing class '%s'", name);
    }
}