package devmgr.api.runtime;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.StandardCharsets;

/**
 * This class implements JSON encoding to an output writer.
 */
public class JsonEncoder
extends Encoder
{
    private Writer m_Out;
    
    /**
     * Constructor.
     * @param writer Writer to which results are sent.
     */
    public 
    JsonEncoder(Writer writer)
    {
        m_Out = writer;
    }

    /**
     * Constructor.
     * @param out Output stream to which results are sent.
     */
    public 
    JsonEncoder(OutputStream out)
    {
        this(new OutputStreamWriter(out, StandardCharsets.UTF_8));
    }
    
    /**
     * Flush the output stream.
     */
    @Override
    protected void
    flush()
    throws EncodeException
    {
        try
        {
            m_Out.flush();
        }
        catch (Exception ex)
        {
            throw EncodeException.build(ex, "Can't flush JSON stream");
        }
    }
    
    /**
     * Getter method for the Writer object to which results are sent.
     * @return
     */
    public Writer
    getWriter()
    throws IOException
    {
        m_Out.flush();
        return m_Out;
    }
    
    /**
     * Emit one or more (non-quoted) strings to the output stream.
     */    
    private void
    emitChar(char c)
    throws EncodeException
    {
        try
        {
            m_Out.append(c);
        }
        catch (Exception ex)
        {
            throw EncodeException.build(ex, "Can't emit char");
        }
    }

    /**
     * Emit one or more (non-quoted) strings to the output stream.
     */    
    private void
    putRawString(String... args)
    throws EncodeException
    {
        try
        {
            for (String s : args)
            {
                m_Out.append(s);
            }
        }
        catch (Exception ex)
        {
            throw EncodeException.build(ex, "Can't put raw string");
        }
    }

    /**
     * Emit a quoted string to the output stream.
     */
    private void
    putQuotedString(String s)
    throws EncodeException
    {
        emitChar('"');
        putRawString(s);
        emitChar('"');
    }

    /**
     * Emit a prefix, which may include a comma to separate this item from
     * its predecessor, and may include a name string followed by a colon.
     * The context value helps us figure out which of these are needed.
     * @param name
     * @param ctx
     */
    private void
    putPrefix(String name, Context ctx)
    throws EncodeException
    {
        // See if we're in an array context.  If so, don't emit a name, and
        // do emit a comma for all but the first element.
        if ( (ctx == Context.ARRAY_FIRST) || (ctx == Context.ARRAY_NONFIRST) )
        {
            if (ctx == Context.ARRAY_NONFIRST)
                emitChar(',');
        }
        
        // If not in array context, emit the quoted name, possibly
        // prefixed with a comma (for non-first items).
        else
        {
             switch (ctx)
            {
                case STRUCT_NONFIRST:
                case UNION_ALTERNATE:
                    emitChar(',');
                    // Fall through...
                    
                case STRUCT_FIRST:
                    putQuotedString(name);
                    emitChar(':');
                    break;
                    
                default:
                    break;
            }
        }
    }
    
    @Override
    protected void 
    encodeByte(byte obj, String name, Context ctx)
    throws EncodeException
    {
        encodeInt((int) obj, name, ctx);
    }
    
    @Override
    protected void 
    encodeShort(short obj, String name, Context ctx)
    throws EncodeException
    {
        encodeInt((int) obj, name, ctx);
    }

    @Override
    protected void 
    encodeInt(int obj, String name, Context ctx)
    throws EncodeException
    {
        putPrefix(name, ctx);
        putRawString(Integer.toString(obj));
    }
    
    @Override
    protected void 
    encodeLong(long obj, String name, Context ctx)
    throws EncodeException
    {
        putPrefix(name, ctx);
        putQuotedString(Long.toString(obj));
    }

    @Override
    protected void 
    encodeBoolean(boolean obj, String name, Context ctx)
    throws EncodeException
    {
        putPrefix(name, ctx);
        putRawString(obj ? "true" : "false");
    }
    
    @Override
    protected void 
    encodeFloat(float obj, String name, Context ctx)
    throws EncodeException
    {
        putPrefix(name, ctx);
        putRawString(Float.toString(obj));
    }
    
    @Override
    protected void 
    encodeDouble(double obj, String name, Context ctx)
    throws EncodeException
    {
        putPrefix(name, ctx);
        putRawString(Double.toString(obj));
    }
    
    @Override
    protected void 
    encodeString(String obj, String name, Context ctx, int maxLength)
    throws EncodeException
    {
        putPrefix(name, ctx);
        putQuotedString(obj);
    }
    
    @Override
    protected void 
    encodeOpaque(String data, String name, Context ctx, OpaqueSpec typeSpec)
    throws EncodeException
    {
        putPrefix(name, ctx);
        putQuotedString(data);
    }

    @Override
    protected void 
    encodeArrayStart(String name, Context ctx, 
                     boolean isVariable, int actualLength)
    throws EncodeException
    {
        putPrefix(name, ctx);
        emitChar('[');
    }
    
    @Override
    protected void 
    encodeArrayEnd(String name, Context ctx, 
                   boolean isVariable, int actualLength)
    throws EncodeException
    {
        emitChar(']');
    }
    
    @Override
    protected void 
    encodeStructStart(String name, Context ctx, StructSpec spec)
    throws EncodeException
    {
        // Do nothing for collapsed structures, since we'll get an
        // ensuing call to encode the one/only field.
        if (spec.isCollapsed)
            return;
        
        putPrefix(name, ctx);
        emitChar('{');
    }
    
    @Override
    protected void 
    encodeStructEnd(String name, Context ctx, StructSpec spec)
    throws EncodeException
    {
        // Do nothing for collapsed structures.
        if (spec.isCollapsed)
            return;
        
        emitChar('}');
    }
    
    @Override
    protected void 
    encodeUnionStart(String name, Context ctx, UnionSpec spec,
                     EnumValue discrimVal)
    throws EncodeException
    {
        putPrefix(name, ctx);
        emitChar('{');
        encodeEnum(discrimVal, spec.discrimName, Context.STRUCT_FIRST);
    }
    
    @Override
    protected void 
    encodeUnionEnd(String name, Context ctx, UnionSpec spec)
    throws EncodeException
    {
        emitChar('}');
    }
    
    @Override
    protected void 
    encodeEnum(EnumValue value, String name, Context ctx)
    throws EncodeException
    {
        putPrefix(name, ctx);
        putQuotedString(value.name);
    }
    
    /**
     * This method is invoked to encode a pointer prefix.  Since JSON
     * encodings do not use such prefixes, we do nothing here.  If the
     * pointer is non-null, we'll get a follow-on request to encode its
     * target object.  Otherwise, null pointers cause no encoding to
     * take place for JSON.
     */
    @Override
    protected void 
    encodePointerPrefix(String name, Context ctx, boolean isNull)
    throws EncodeException
    {
    }
}
