package devmgr.api.runtime;

import devmgr.api.symbolutils.ByteUtils;

import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.util.LinkedList;

/**
 * This class implements an XDR decoder.  It is constructed using an
 * XDRMessageIOV object that supplies the raw data to be extracted.
 * As part of the extraction process, the data elements are translated 
 * from network-encoded form to the local representation.
 */
public class XdrDecoder
extends Decoder
{
    private static final int XDR_UNIT_SIZE = 4;

    private XdrMessage m_Msg;

    private int m_TotalMsgSize;

    private int m_OffsetInMsg;
    private int m_NextBuf;
    private byte[] m_Buf;
    private int m_BufSize;
    private int m_OffsetInBuf;
    
    private LinkedList<Object> m_ContextList =
        new LinkedList<Object>();
    
    private int m_NestingDepth = 0;
    private boolean m_Verbose = false;

    private static class SavedPosition
    {
        int offsetInMsg;
        int nextBuf;
        byte[] buf;
        int bufSize;
        int offsetInBuf;
    }

    private static class ArrayContext
    {
        int count;
        int consumed;
    }
    
    private static class StructContext
    {
        int encodedLength;
        int startPos;
        int nextField;
    }
    
    private static class UnionContext
    {
        int encodedLength;
        int startPos;
        TypeBinding altBinding;
        boolean isDone;
    }
    
    /**
     * Constructor for XDRInputStream objects.  Such objects must be
     * initialized with an existing message, which is in the form of
     * an XDRMessageIOV object.
     * @param rawMsg Message with raw input data
     */
    public XdrDecoder(XdrMessage rawMsg)
    {
        m_Msg = rawMsg;
    
        m_TotalMsgSize = m_Msg.getMessageSize();
        m_OffsetInMsg = 0;
        m_NextBuf = 0;
    
        m_Buf = null;
        m_BufSize = 0;
        m_OffsetInBuf = 0;
    
        return;
    }

    /**
     * Sets the verbose/quiet mode.  When in verbose mode, state/debug
     * information is written to System.err.
     * @param verbose True for verbose output; false for quiet mode.
     */
    public void
    setVerbose(boolean verbose)
    {
        m_Verbose = verbose;
    }
    
    /**
     * Show state/debug information.
     * @param fmt Format string to use.
     * @param args Format arguments.
     */
    private void 
    showState(String fmt, Object... args)
    {
        if (! m_Verbose)
            return;
        
        System.err.print("XDR-D:");
        for (int i = 0; i < m_NestingDepth; i++)
            System.err.print("  ");
        
        System.err.printf(fmt, args);
        
        System.err.printf(" [pos=%d]%n", m_OffsetInMsg);
    }
    
    /**
     * Returns the current position within the input message.
     * @return Current position.
     */
    private SavedPosition getPosition()
    {
        SavedPosition pos = new SavedPosition();
        
        pos.offsetInMsg = m_OffsetInMsg;
        pos.nextBuf = m_NextBuf;
        pos.buf = m_Buf;
        pos.bufSize = m_BufSize;
        pos.offsetInBuf = m_OffsetInBuf;
        
        return pos;
    }

    /**
     * Sets the position for further decoding operations to the specified
     * point in the input message.  
     * @param pos Desired position, obtained in a prior call to getPosition().
     */
    private void setPosition(SavedPosition pos)
    {
        m_OffsetInMsg = pos.offsetInMsg;
        m_NextBuf = pos.nextBuf;
        m_Buf = pos.buf;
        m_BufSize = pos.bufSize;
        m_OffsetInBuf = pos.offsetInBuf;
    }

    /**
     * Skips over the specified number of bytes in the input stream.
     * @param length Number of bytes to skip.
     * @exception DecodeException If any input stream failure occurs.
     */
    private void skip(int length)
    throws DecodeException    
    {
        for (int i = 0; i < length; i++)
            getByte();
    }

    /**
     * Fetch the next byte value from the raw message data.
     * @return The next raw byte value from the input stream.
     * @exception DecodeException If any input stream failure occurs.
     */
    private byte getByte() throws DecodeException
    {
        if (m_OffsetInBuf >= m_BufSize)
        {
            if (m_OffsetInMsg >= m_TotalMsgSize)
            {
                throw new DecodeException("XDR: buffer underflow");
            }

            m_Buf = m_Msg.getBuffer(m_NextBuf++);
            m_OffsetInBuf = 0;
            m_BufSize = Math.min(m_Buf.length, 
                                 m_TotalMsgSize - m_OffsetInMsg);
        }

        m_OffsetInMsg++;
        return m_Buf[m_OffsetInBuf++];
    }

    /**
     * Determines the number of pad bytes associated with the
     * specified data length to obtain an even multiple of the
     * XDR_UNIT_SIZE value.
     * @param amt Amount of data used in operation
     * @return The amount of padding needed for the specified byte count.
     */
    private int calculatePad(int amt)
    {
        int residual = amt % XDR_UNIT_SIZE;

        if (residual == 0)
            return 0;
        else
            return XDR_UNIT_SIZE - residual;
    }

    /**
     * Decodes and returns a (32-bit) integer from the XDR stream.
     * @return The next integer data element in the stream.
     * @exception DecodeException If any input stream failure occurs.
     */
    private int fetchInt() 
    throws DecodeException
    {
        int b1 = (int)(getByte()) & 0xFF;
        int b2 = (int)(getByte()) & 0xFF;
        int b3 = (int)(getByte()) & 0xFF;
        int b4 = (int)(getByte()) & 0xFF;
    
        return (b1 << 24) |
               (b2 << 16) |
               (b3 <<  8) |
               (b4 <<  0) ;
    }

    /**
     * Decodes and returns a (64-bit) integer from the XDR stream.
     * @return The next long data element in the stream.
     * @exception DecodeException If any input stream failure occurs.
     */
    private Long fetchLong()
    throws DecodeException
    {
        return ( ((long) fetchInt()) << 32) |
                 (((long) fetchInt()) & 0xFFFFFFFFL);
    }

    /**
     * Fetches a string of the specified length from the XDR input
     * stream and returns it to the caller.  The caller provides the
     * previously-determined string length as an input parameter.
     * This routine also eats any padding bytes that are present in
     * the stream after the string data proper.
     * @param length Length of the string to be fetched.
     * @return A string of the specified size constructed from input data.
     * @exception DecodeException If any input stream failure occurs.
     */
    private String fetchString(int length) 
    throws DecodeException
    {
        StringBuilder sb = new StringBuilder(length);
        int pad;
        int i;

        for (i = 0; i < length; i++)
            sb.append((char) getByte());

        pad = calculatePad(length);
        for (i = 0; i < pad; i++)
             getByte();

        return sb.toString();
    }
    
    /**
     * This method gets the next name (discriminator OR data field) in
     * the input stream for the specified union.  If we reach the end
     * of the union, a null value is returned.
     */
    @Override
    protected String 
    getUnionItem(String enclosingName, Context ctx, UnionSpec spec)
    throws DecodeException
    {
        UnionContext decCtx = (UnionContext) m_ContextList.getLast();

        if ( spec.isVarLen &&
            (m_OffsetInMsg >= (decCtx.startPos + decCtx.encodedLength)) )
           return null;
        
        if (ctx == Context.UNION_FIRST)
        {
            String name = spec.discrimName;
            
            // Peek at the discriminator to determine the binding
            // for the alternate.  We'll need this on the ensuing
            // call to this method.
            SavedPosition savedPos = getPosition();
            EnumValue discrim = 
                decodeEnum(name, (EnumSpec) spec.discrimType);
            setPosition(savedPos);
            
            decCtx.altBinding = spec.caseMap.get(discrim.value);
            if (decCtx.altBinding == null)
                decCtx.altBinding = spec.defaultBinding;
            
            return name;
        }
        
        // This code handles requests AFTER the first one, which should
        // always be for the alternate value.
        else
        {
            if (decCtx.isDone)
                return null;
            
            if (decCtx.altBinding == null)
                return null;
            
            decCtx.isDone = true;
            return decCtx.altBinding.itemName;
        }
    }

    /**
     * This method gets the type binding information for the next
     * field in the input stream for the specified structure.  If we
     * reach the end of the structure, a null value is returned.
     */
    @Override
    protected TypeBinding
    getStructItem(String enclosingName, Context ctx, StructSpec spec)
    throws DecodeException
    {
        StructContext decCtx = (StructContext) m_ContextList.getLast();

        if ( spec.isVarLen &&
             (m_OffsetInMsg >= (decCtx.startPos + decCtx.encodedLength)) )
            return null;
        
        if (decCtx.nextField >= spec.fields.length)
            return null;
        
        return spec.fields[decCtx.nextField++];
    }

    @Override
    protected boolean
    isEndOfArray(String name, Context ctx)
    throws DecodeException
    {
        ArrayContext decCtx = (ArrayContext) m_ContextList.getLast();
        
        if (decCtx.consumed >= decCtx.count)
            return true;
        
        decCtx.consumed++;
        return false;
    }

    @Override
    protected byte decodeByte(String name)
    throws DecodeException
    {
        return (byte) fetchInt();
    }

    @Override
    protected short decodeShort(String name)
    throws DecodeException
    {
        return (short) fetchInt();
    }

    @Override
    protected int decodeInt(String name)
    throws DecodeException
    {
        return fetchInt();
    }

    @Override
    protected long decodeLong(String name)
    throws DecodeException
    {
        return fetchLong();
    }

    @Override
    protected boolean decodeBoolean(String name)
    throws DecodeException
    {
        int value = fetchInt();
        
        if (value == 0)
            return false;
        else if (value == 1)
            return true;
        
        throw new DecodeException("XDR: bad boolean: "+ name);
    }

    @Override
    protected float decodeFloat(String name)
    throws DecodeException
    {
        return Float.intBitsToFloat(fetchInt());
    }

    @Override
    protected double decodeDouble(String name)
    throws DecodeException
    {
        return Double.longBitsToDouble(fetchLong());
    }

    @Override
    protected String decodeString(String name, int maxLength)
    throws DecodeException
    {
        int length = fetchInt();
        
        if (length > maxLength)
            throw new DecodeException("XDR: string overflow");
        
        return fetchString(length);
    }

    @Override
    protected String 
    decodeOpaque(String name, OpaqueSpec typeSpec)
    throws DecodeException
    {
        String result = null;
        
        int length = (typeSpec.isVariable ? 
            fetchInt() : typeSpec.maxLength);

        if (length > typeSpec.maxLength)
            throw new DecodeException("XDR: opaque too long");

        switch (typeSpec.dataType)
        {
            case BINARY:
                byte[] buf = new byte[length];
                for (int i = 0; i < length; i++)
                    buf[i] = getByte();
                
                result = ApiUtil.opaqueToString(buf);
                break;
                
            case DBCS:
                if ((length % 2) != 0) {

                    //throw new DecodeException("XDR: odd DBCS opaque length");
                    StringBuilder dbcs = new StringBuilder();
                    byte[] dataToDebug=new byte[length-1];
                    for (int i=0;i<dataToDebug.length;i++){
                        dataToDebug[i]=getByte();
                    }
                    getByte();//read in the odd bird, probably a null terminator

                    for (int i = 0; i < dataToDebug.length ; i+=2) {
                        int c = ((((int) dataToDebug[i]) << 8) |
                                (((int) dataToDebug[i+1]) & 0xFF)) & 0xFFFF;

                        dbcs.append((char) c);
                    }

                    result = dbcs.toString();


                }else {

                    StringBuilder dbcs = new StringBuilder();
                    for (int i = 0; i < length / 2; i++) {
                        int c = ((((int) getByte()) << 8) |
                                (((int) getByte()) & 0xFF)) & 0xFFFF;

                        dbcs.append((char) c);
                    }

                    result = dbcs.toString();
                }
                break;
                
            case SBCS:
                StringBuilder sbcs = new StringBuilder();
                boolean terminated = false;
                for (int i = 0; i < length; i++)
                {
                    int c = ((int) getByte()) & 0xFF;
                    if (!terminated) {
                        if (c != 0) {
                            sbcs.append((char) c);
                        } else {
                            terminated = true;
                        }
                    }
                }

                result = sbcs.toString();
                break;
            case IPV4ADDR:
                if (length != 4)
                    throw new DecodeException("XDR: odd IPV4ADDR opaque length");

                byte[] data=new byte[4];
                data[0]=getByte();
                data[1]=getByte();
                data[2]=getByte();
                data[3]=getByte();
                result=ByteUtils.getIpAddressString(data);
                break;

        }
        
        // Skip over any padding bytes.
        int pad = calculatePad(length);
        for (int i = 0; i < pad; i++)
            getByte();
        
        return result;
    }

    @Override
    protected void 
    decodeArrayStart(String name, Context ctx,
                     boolean isVariable, int maxLength)
    throws DecodeException
    {
        ArrayContext newCtx = new ArrayContext();
        
        if (isVariable)
        {
            newCtx.count = fetchInt();
            if ( (maxLength != 0) && (newCtx.count > maxLength) )
                throw new DecodeException("XDR: variable array overflow");
        }
        else
        {
            newCtx.count = maxLength;
        }
        
        newCtx.consumed = 0;
        m_ContextList.add(newCtx);
    }

    @Override
    protected void 
    decodeArrayEnd(String name, boolean isVariable, int maxLength)
    throws DecodeException
    {
        m_ContextList.removeLast();
    }

    @Override
    protected boolean 
    decodeStructStart(String name, Context ctx, StructSpec spec)
    throws DecodeException
    {
        showState("+struct: %s", name);
        m_NestingDepth++;
        
        // Create a new context for processing the structure.
        StructContext newCtx = new StructContext();
        
        if (spec.isVarLen)
            newCtx.encodedLength = fetchInt();
        
        newCtx.nextField = 0;
        newCtx.startPos = m_OffsetInMsg;
        m_ContextList.add(newCtx);
        
        return ( spec.isVarLen ? (newCtx.encodedLength > 0) : true );
    }

    @Override
    protected void 
    decodeStructEnd(String name, StructSpec spec)
    throws DecodeException
    {
        m_NestingDepth--;
        showState("-struct: %s", name);

        StructContext decCtx = (StructContext) m_ContextList.removeLast();
        
        if (! spec.isVarLen)
            return;

        int slop = decCtx.startPos + decCtx.encodedLength - m_OffsetInMsg;
        
        if (slop < 0)
            throw DecodeException.build("XDR: misaligned structure: %s", name);
        else if (slop > 0)
            skip(slop);
    }

    @Override
    protected boolean 
    decodeUnionStart(String name, Context ctx, UnionSpec spec)
    throws DecodeException
    {
        showState("+union: %s", name);
        m_NestingDepth++;

        // Create a new context for processing the union.
        UnionContext newCtx = new UnionContext();

        if (spec.isVarLen)
            newCtx.encodedLength = fetchInt();
        
        newCtx.startPos = m_OffsetInMsg;
        newCtx.isDone = false;
        m_ContextList.add(newCtx);
        
        return ( spec.isVarLen ? (newCtx.encodedLength > 0) : true );
    }

    @Override
    protected void 
    decodeUnionEnd(String name, UnionSpec spec)
    throws DecodeException
    {
        m_NestingDepth--;
        showState("-union: %s", name);
        
        UnionContext decCtx = (UnionContext) m_ContextList.removeLast();
        
        if (! spec.isVarLen)
            return;

        int slop = decCtx.startPos + decCtx.encodedLength - m_OffsetInMsg;
        
        if (slop < 0)
            throw DecodeException.build("XDR: misaligned union: %s", name);
        else if (slop > 0)
            skip(slop);
    }

    @Override
    protected EnumValue 
    decodeEnum(String name, EnumSpec typeSpec)
    throws DecodeException
    {
        int value = fetchInt();
        String label = typeSpec.valueToNameMap.get(value);

        // If we encounter an enumeration value that is simply not
        // defined for this type, we need to force it to a legal
        // value.  That's what the "BOGUS" stuff is for.
        if (label == null)
        {
            label = ApiUtil.BOGUS_ENUM_NAME;
            value = ApiUtil.BOGUS_ENUM_VALUE;
        }
        
        return new EnumValue(label, value);
    }
    
    protected EnumValue[] 
    decodeEnumSet(String name, Context ctx, EnumSetSpec spec)
    throws DecodeException
    {
        EnumSpec enumType = spec.enumType;
        int bits = fetchInt();
        LinkedList<EnumValue> values = new LinkedList<EnumValue>();
        
        for (int mask : enumType.valueToNameMap.keySet())
        {
            // If all (remaining) bits are zero, we're done.
            if (bits == 0)
                break;
            
            // Ignore zero values.
            if (mask == 0)
                continue;
            
            // Ignore the "bogus" value representation.
            if (mask == ApiUtil.BOGUS_ENUM_VALUE)
                continue;
           
            // See if this mask value is present.
            if ( (bits & mask) == mask )
            {
                // Add to our accumulator list.
                values.add(new EnumValue(enumType.valueToNameMap.get(mask), 
                                         mask));
                
                // Exclude this mask's value from further work.
                bits &= (~ mask);
            }
        }

        EnumValue[] result = new EnumValue[values.size()];
        return values.toArray(result);
    }
}
