/**************************************************************************
 *                                                                         *
 *  Copyright 1998-2011 NetApp, Inc. All Rights Reserved.                  *
 *                                                                         *
 *  This file is confidential and a trade secret of NetApp, Inc.  The      *
 *  receipt of or possession of this file does not convey any rights to    *
 *  reproduce or disclose its contents or to manufacture, use, or sell     *
 *  anything it may describe, in whole, or in part, without the specific   *
 *  written consent of NetApp, Inc.                                        *
 *                                                                         *
 **************************************************************************/

package devmgr.api.security;

import devmgr.api.symbolutils.ByteUtils;

import java.io.UnsupportedEncodingException;
import java.security.*;

/**
 * VERSION: 3 <BR>
 * UPDATE DATE: Wed Mar 10 07:17:16 1999 <BR>
 * <BR>
 * CONTACTS: <BR>
 * Originator: Bill Delaney <BR>
 * Updater: bdelaney <BR>
 * <BR>
 * DESCRIPTION: <BR>
 * This class implements Passphrase-Based Encryption as specified in
 * the PKCS#5 Version 2.0 standard published by RSA Laboratories.
 * <BR><BR>
 * NOTES/ASSUMPTIONS: <BR>
 * This implementation always uses SHA-1 as the message digest for
 * generating the 56-bit DES key and the 8-byte Initialization Vector.
 *
 * @deprecated This class uses the weakest encryption supported by the array. Use EncryptedPasswordFactory instead
 *          to automatically use the strongest encryption supported by the array.
 */
@Deprecated
public class PBE
{
    private static final int BLOCK_SIZE = DESWithCBC.BLOCK_SIZE;

    private DESWithCBC m_DesEngine = null;
    private boolean m_Encrypting;

    /**
     * DESCRIPTION: <BR>
     * This method initializes the PBE object for all ensuing
     * operations.  The password and salt attributes are specified
     * by the caller, along with the iteration count to be used
     * to generate the DES key and Initialization Vector (IV).
     * <BR><BR>
     * NOTES/ASSUMPTIONS: <BR>
     * None.
     * <BR>
     * @param encrypt True means encrypt; false means decrypt
     * @param password Buffer with password data (raw bytes)
     * @param salt The 8-byte salt value for generating key and IV
     * @param iterations Digest iterations for generating key and IV
     * @exception NoSuchAlgorithmException SHA-1 digest not available
     */
    public void init(boolean encrypt, byte[] password,
                     byte[] salt, int iterations)
            throws NoSuchAlgorithmException
    {
        byte[] digest;
        byte[] key;
        byte[] iv;
        MessageDigest sha = MessageDigest.getInstance("SHA");

      /*
       * Keep track of the encryption/decryption direction.
       */
        m_Encrypting = encrypt;

      /*
       * Feed the password and salt into the digest engine to
       * complete the first iteration.
       */
        sha.update(password);
        sha.update(salt, 0, 8);

      /*
       * For each ensuing iteration, feed in the digest value
       * from the previous iteration.
       */
        for (int i = 1; i < iterations; i++)
        {
            digest = sha.digest();
            sha.update(digest);
        }

      /*
       * Extract the final digest value and use it to determine
       * the key and initialization vector (IV) for the DES engine.
       */
        digest = sha.digest();

        key = new byte[8];
        iv = new byte[8];
        System.arraycopy(digest, 0, key, 0, 8);
        System.arraycopy(digest, digest.length - 8, iv, 0, 8);

        m_DesEngine = new DESWithCBC(key, encrypt, iv);

        return;
    }

    /**
     * Function for backwards compatible
     * @param input
     * @return
     * @throws InvalidParameterException
     */
    public byte[] crypt(String input) throws InvalidParameterException{
        byte[] bytes;
        try {
            bytes = input.getBytes(ByteUtils.getUnicodeEncoding());
        } catch (UnsupportedEncodingException e) {
            throw new InvalidParameterException(e.getLocalizedMessage());
        }
        return crypt(bytes);

    }
    /**
     * DESCRIPTION: <BR>
     * This method processes the specified buffer and returns the
     * corresponding encrypted or decrypted data.
     * <BR><BR>
     * NOTES/ASSUMPTIONS: <BR>
     * This method can only be invoked once after initializing a
     * PBE context.  Reinitialization is required before any ensuing
     * operations are allowed.
     * <BR>
     * @param input The input data to be processed (a byte array)
     * @exception InvalidParameterException Invalid data given for decryption
     * @return The corresponding output data (a byte array)
     */
    public byte[] crypt(byte[] input)
            throws InvalidParameterException
    {

        int amtDone = 0;
        int outCount = 0;
        byte padValue;
        byte[] inBlock = new byte[BLOCK_SIZE];
        byte[] outBlock = new byte[BLOCK_SIZE];
        byte[] output;
        int outSize;

      /*
       * Check for bogus input sizes and give up if we find one.
       */
        if ( (! m_Encrypting) && ((input.length % 8) != 0) )
        {
            throw new InvalidParameterException("Invalid input buffer size");
        }

      /*
       * Determine the output buffer size.  If we're encrypting, we
       * know the value for certain.  If we're decrypting, we may
       * need to reallocate to get rid of pad bytes before returning
       * to the caller.
       */
        if (m_Encrypting)
        {
            outSize = input.length + (8 - (input.length % 8));
        }
        else
        {
            outSize = input.length;
        }

      /*
       * Create an output buffer in which to accumulate result data.
       */
        output = new byte[outSize];

      /*
       * Process the input one block at a time, putting processed
       * data into our newly-created output buffer.
       */
        while ( (input.length - amtDone) >= BLOCK_SIZE )
        {
            System.arraycopy(input, amtDone, inBlock, 0, BLOCK_SIZE);
            m_DesEngine.crypt(inBlock, outBlock);
            System.arraycopy(outBlock, 0, output, outCount, BLOCK_SIZE);

            amtDone += BLOCK_SIZE;
            outCount += BLOCK_SIZE;
        }

      /*
       * If we're encrypting, we need to generate pad bytes.  The value
       * used for each pad byte is the number of pad bytes that are
       * being used to reach a multiple of DES_BLOCK_SIZE in total size.
       * We construct the pad block (which may start with residual data
       * from the input buffer) in a special local buffer so we can
       * encrypt it and add the result to the caller's output buffer.
       */
        if (m_Encrypting)
        {
            padValue = (byte) (BLOCK_SIZE - (input.length - amtDone));

            System.arraycopy(input, amtDone, inBlock, 0, input.length - amtDone);
            for (int i = input.length - amtDone; i < BLOCK_SIZE; i++)
            {
                inBlock[i] = padValue;
            }

            m_DesEngine.crypt(inBlock, outBlock);
            System.arraycopy(outBlock, 0, output, outCount, BLOCK_SIZE);
        }

      /*
       *  Otherwise, we must be decrypting, so we validate and then strip
       *  out the extra padding bytes so the caller doesn't have to worry
       *  about them.
       */
        else
        {
         /*
          * Extract the pad count from the last byte of the decrypted
          * data.  Make sure it's a legitimate value (in the range 1-8).
          */
            padValue = output[output.length - 1];

            if ( (padValue < 1) || (padValue > BLOCK_SIZE) )
            {
                throw new InvalidParameterException("Invalid pad value in input");
            }

         /*
          * Make sure all padding bytes have the same value.  If not,
          * consider the data to be invalid and throw an error back to
          * the caller.
          */
            for (int i = 1; i < padValue; i++)
            {
                if (output[output.length - 1 - i] != padValue)
                {
                    throw new InvalidParameterException("Mismatched pad bytes in input");
                }
            }

         /*
          * The padding looks legitimate, so strip it by allocating
          * a new buffer of the true length and using it for return
          * to the caller.
          */
            byte[] newOutput = new byte[output.length - padValue];
            System.arraycopy(output, 0, newOutput, 0, newOutput.length);
            output = newOutput;
        }

        return output;
    }
}
