package devmgr.api.symbol;

import devmgr.api.runtime.*;
import java.io.Serializable;
import javax.xml.bind.annotation.*;

/** This enumeration object is used to represent one of a set of possible functional capabilities supported by the RAID controller firmware. */
@XmlRootElement
@io.swagger.annotations.ApiModel()
@TypeTableIndex(160)
public enum Capability implements Serializable {
		/** This capability is intended to be used in cases where a feature identified by a feature ID has no corresponding capability code. */
		none(0),
		/** The shared volume capability is supported. When this capability is present, all controllers within the storage array have equal, concurrent access to any given volume. Otherwise, each defined volume is owned by exactly one controller at a time. */
		sharedVolume(1),
		/** The storage pool capability is supported for up to 4 storage partition definitions. */
		storagePoolsTo4(2),
		/** Mixed raid level support within groups */
		mixedRaidlevel(3),
		/** The automatic code synchronization capability is supported and active. This causes controllers to automatically synchronize their firmware levels to ensure that a multi-controller configuration always uses compatible firmware on all controllers. */
		autoCodeSync(4),
		/** The automatic LUN transfer capability is supported and active. This causes a controller to automatically assume ownership (from the alternate controller) of an unowned volume group if the host system attempts to access any volume within the volume group. */
		autoLunTransfer(5),
		/** The sub-LUNs allowed capability is supported and active. When this capability is present, the controller allows volume creation on free extents within an existing volume group. Consequently, multiple volumes can be created per volume group. If this capability is not present, then volume groups are limited to one volume each. */
		subLunsAllowed(6),
		/** The storage pool capability is supported for up to 8 storage partition definitions. */
		storagePoolsTo8(7),
		/** The storage pool capability is supported for up to 2 storage partition definitions. When this capability is present, a more generalized I/O path mapping is supported for accessing individual volumes within the array. Without it, a given volume must be addressed using a single, unique logical unit number, regardless of the I/O host that initiates the operation or the I/O channel that is used to perform the volume access. If this capability is not supported, a volume's logical unit number (for I/O access) is determined by the volumeHandle field of the associated Volume object. */
		storagePoolsTo2(8),
		/** The storage pool capability is supported for up to the maximum storage partition definitions supported by the controller firmware. */
		storagePoolsToMax(9),
		/** The storage pool capability is supported for up to 64 storage partition definitions. */
		storagePoolsTo64(10),
		/** The storage pool capability is supported for up to 16 storage partition definitions. */
		storagePoolsTo16(11),
		/** The snapshot copy feature is supported. The Snapshot Copy feature allows the end-user to quickly create a single point-in time image or "snapshot" of a volume. */
		snapshots(12),
		/** The remote mirroring feature copy is supported. The Remote Mirroring feature allows the end-user to mirror data from one site to remote site some distance away. */
		remoteMirroring(13),
		/** The volume copy feature is supported. Volume copy allows for replicating volume data within a storage array. */
		volumeCopy(14),
		/** The staged online firmware download feature is supported. */
		stagedDownload(15),
		/** The controller supports the mixing of drive types within the array. Depending on the NVSRAM configuration of the controller, Mixed Drive Types may be either a premium feature or a capability that is always available. */
		mixedDriveTypes(16),
		/** The controller is configured (via NVSRAM) as being subject to the Gold Key feature. When the controller is subject to the Gold Key feature and when the Gold Key is enabled, other premium features such as Remote Volume Mirroring, Volume Copy, and Mixed Drive Types (MDT) may also be enabled on the storage array. */
		goldKey(17),
		/** The ability to add more drive trays is supported up to a specific limit. */
		driveTrayExpansion(18),
		/** The ability to migrate a storage array from one feature bundle to another. */
		bundleMigration(19),
		/** The storage pool capability is supported for up to 128 storage partition definitions. */
		storagePoolsTo128(20),
		/** The storage pool capability is supported for up to 256 storage partition definitions. */
		storagePoolsTo256(21),
		/** The ability to create and manage RAID 6 volumes. */
		raid6(22),
		/** The ability to modify the array performance level via a feature key. */
		performanceTier(23),
		/** The storage pool capability is supported for up to 32 storage partition definitions. */
		storagePoolsTo32(24),
		/** The storage pool capability is supported for up to 96 storage partition definitions. */
		storagePoolsTo96(25),
		/** The storage pool capability is supported for up to 192 storage partition definitions. */
		storagePoolsTo192(26),
		/** The storage pool capability is supported for up to 512 storage partition definitions. */
		storagePoolsTo512(27),
		/** The remote mirroring capability is supported for up to 16 remote mirrors. */
		remoteMirrorsTo16(28),
		/** The remote mirroring capability is supported for up to 32 remote mirrors. */
		remoteMirrorsTo32(29),
		/** The remote mirroring capability is supported for up to 64 remote mirrors. */
		remoteMirrorsTo64(30),
		/** The remote mirroring capability is supported for up to 128 remote mirrors. */
		remoteMirrorsTo128(31),
		/** The snapshots-per-volume capability is supported for up to 4 snapshots per volume. */
		snapshotsPerVolTo4(32),
		/** The snapshots-per-volume capability is supported for up to 8 snapshots per volume. */
		snapshotsPerVolTo8(33),
		/** The snapshots-per-volume capability is supported for up to 16 snapshots per volume. */
		snapshotsPerVolTo16(34),
		/** The snapshots-per-volume capability is supported for up to 2 snapshots per volume. */
		snapshotsPerVolTo2(35),
		/** The Secure Volume capability is supported. Volumes can reside within secure volume groups. */
		secureVolume(36),
		/** This capability is indicated if the storage array is protection information capable, even if the protection information feature is currently disabled. */
		protectionInformation(37),
		/** Solid State Disks are supported. */
		ssdSupport(38),
		/** The storage array is capable of supporting up to 112 drive slots. */
		driveSlotLimitTo112(45),
		/** The storage array is capable of supporting up to 120 drive slots. */
		driveSlotLimitTo120(46),
		/** The storage array is capable of supporting up to 256 drive slots. */
		driveSlotLimitTo256(47),
		/** The storage array is capable of supporting up to 448 drive slots. */
		driveSlotLimitTo448(48),
		/** The storage array is capable of supporting up to 480 drive slots. */
		driveSlotLimitTo480(49),
		/** This enumeration value is not used. */
		driveSlotLimitToMax(50),
		/** The storage array is capable of supporting up to the specified number of drive slots. This feature is controlled by Type 2 Licensing. */
		driveSlotLimit(51),
		/** The storage array is capable of supporting up to 12 drive slots. */
		driveSlotLimitTo12(53),
		/** The storage array is capable of supporting up to 16 drive slots. */
		driveSlotLimitTo16(54),
		/** The storage array is capable of supporting up to 24 drive slots. */
		driveSlotLimitTo24(55),
		/** The storage array is capable of supporting up to 32 drive slots. */
		driveSlotLimitTo32(56),
		/** The storage array is capable of supporting up to 48 drive slots. */
		driveSlotLimitTo48(57),
		/** The storage array is capable of supporting up to 60 drive slots. */
		driveSlotLimitTo60(58),
		/** The storage array is capable of supporting up to 64 drive slots. */
		driveSlotLimitTo64(59),
		/** The storage array is capable of supporting up to 72 drive slots. */
		driveSlotLimitTo72(60),
		/** The storage array is capable of supporting up to 96 drive slots. */
		driveSlotLimitTo96(61),
		/** The storage array is capable of supporting up to 128 drive slots. */
		driveSlotLimitTo128(62),
		/** The storage array is capable of supporting up to 136 drive slots. */
		driveSlotLimitTo136(63),
		/** The storage array is capable of supporting up to 144 drive slots. */
		driveSlotLimitTo144(64),
		/** The storage array is capable of supporting up to 180 drive slots. */
		driveSlotLimitTo180(65),
		/** The storage array is capable of supporting up to 192 drive slots. */
		driveSlotLimitTo192(66),
		/** The storage array is capable of supporting up to 272 drive slots. */
		driveSlotLimitTo272(67),
		/** The storage array is capable of supporting FDE proxy key management. */
		fdeProxyKeyManagement(68),
		/** The remote mirroring capability is supported for up to 8 remote mirrors. */
		remoteMirrorsTo8(84),
		/** The storage array is capable of supporting up to 384 drive slots. */
		driveSlotLimitTo384(93),
		/** The storage array is capable of supporting up to 300 drive slots. */
		driveSlotLimitTo300(94),
		/** The storage array is capable of supporting up to 360 drive slots. */
		driveSlotLimitTo360(95),
		/** The storage array is capable of supporting the Flash Read Cache feature. This feature is controlled by Type 2 Licensing. */
		flashReadCache(96),
		/** Storage Pools are supported. This feature is controlled by Type 2 Licensing. */
		storagePoolsType2(105),
		/** Remote Volume Mirroring is supported. This feature is controlled by Type 2 Licensing. */
		remoteMirroringType2(107),
		/** Checkpoint-based Async Remote Volume Mirroring is supported. This feature is controlled by Type 2 Licensing. The limit controlled by this key is the total number of Async Mirrors allowed per array. */
		totalNumberOfArvmMirrorsPerArray(109),
		/** PiT Groups are supported. This feature is controlled by Type 2 Licensing. The limit controlled by this key is the total number of PiTs per Array. */
		totalNumberOfPitsPerArray(110),
		/** Thin Volume Provisioning is supported. This feature is controlled by Type 2 Licensing. The limit controlled by this key is the total number of Thin Volumes per Array. */
		totalNumberOfThinVolumesPerArray(111),
		/** The storage array is capable of supporting up to 240 drive slots. */
		driveSlotLimitTo240(113),
		/** Snapshots are supported. This feature is controlled by Type 2 Licensing. */
		snapshotsType2(114),
		/** Capability code for a premium feature key to enable the feature id FEATURE_ID_TARGET_PORT_LUN_MAPPING. */
		targetPortLunMapping(115),
		__UNDEFINED(464378545);

		public final int value;

		private Capability(int v) {
			this.value = v;
		}
}
