/**************************************************************************
 *                                                                         *
 *  Copyright 2001-2011 NetApp, Inc. All Rights Reserved.                  *
 *                                                                         *
 *  This file is confidential and a trade secret of NetApp, Inc.  The      *
 *  receipt of or possession of this file does not convey any rights to    *
 *  reproduce or disclose its contents or to manufacture, use, or sell     *
 *  anything it may describe, in whole, or in part, without the specific   *
 *  written consent of NetApp, Inc.                                        *
 *                                                                         *
 **************************************************************************/

package devmgr.api.symbolutils;

import java.security.MessageDigest;

import devmgr.api.rpc.RPCError;
import devmgr.api.rpc.RPCError.ErrorType;
import devmgr.api.runtime.ApiDefinition;
import devmgr.api.runtime.EncodeException;
import devmgr.api.runtime.TypeSpec;
import devmgr.api.runtime.XdrEncoder;
import devmgr.api.runtime.XdrMessage;
import devmgr.api.symbol.RemoteArrayAuthentication;
import devmgr.api.symbol.RemoteMirrorAuthentication;
import devmgr.api.symbol.SAIdentifier;


/**

CONTACTS:                                                               <BR>
Originator:   Juan Gatica                                               <BR>

DESCRIPTION:                                                            <BR>
Convenience class that creates the digest needed for authenticating
with remote array in an Establish Mirror operation.
<P>

NOTES/ASSUMPTIONS:                                                      <BR>
None.
<P>

REFERENCE:                                                              <BR>
List of reference documents - typically will be a list of applicable
software design notes; a software design note reference should be in
the form file-name, title,
e.g., PRD001A.DOC, Notes on Portable RDAC Prototype

*/

public class SYMbolRVMAuthGen
{
   private static final int DIGEST_LENGTH = 20;

   /**
   Method to encode the remote array's password for secure transmission
   to the local array.  This method is used in RVM establish mirror
   operations.   <P>

   NOTES/ASSUMPTIONS: <BR>
   None.    <P>

   @param  said The Storage Array Identifier of the local array.
   @param  remotePassword The password for the remote array.
   @param  configGeneration The config generation number for the local array
   @param  remoteVolumeWWN The WWN of the remote array's volume to use
   @return  Array of bytes with the message digest.
   @exception ManagementOperationFailedException  Thrown if the required encryption classes are not present .
   */
   public static byte[] encodeRemoteAuthenticator(ApiDefinition apiDef, SAIdentifier said, String remotePassword,
                              long configGeneration, String remoteVolumeWWN)
                                                   throws RPCError
   {
      RemoteMirrorAuthentication payload = new RemoteMirrorAuthentication();
      byte                       destination[] = new byte[DIGEST_LENGTH];
      MessageDigest   sha;
      XdrEncoder encoder = new XdrEncoder();
      XdrMessage  msg;

      byte            password[] = getBytes(remotePassword);
      int             totalSize = 0;
      int             amtDone = 0;
      int             nextBuf = 0;
      int             amt;
      int             offset;
      byte[]          result;
      byte[]          buf;
      
      TypeSpec authTypeSpec = apiDef.getTypeSpec(RemoteMirrorAuthentication.class);
      
      // Build remote authenticator
      
      payload.setArrayId(said);
      payload.setCfgGen(configGeneration);
      payload.setVolWwn(remoteVolumeWWN);
            
      try
      {
         sha = MessageDigest.getInstance("SHA");
      }
      catch (Exception ex)
      {
         throw new RPCError(ErrorType.AUTH_ALGORITHM);
      }
      
      sha.update(password);
      
      try
      {
         encoder.encode(payload, authTypeSpec);
      }
      catch (EncodeException e)
      {
         throw new RPCError(ErrorType.AUTH_ALGORITHM);
      }
      
      msg = encoder.getMessage();
      totalSize = msg.getMessageSize();

      /*
       * Iteratively update the digest output value by feeding the
       * digest engine with each buffer that comprises the message.
       */
      while (amtDone < totalSize)
      {
         buf = msg.getBuffer(nextBuf);
         amt = Math.min(buf.length, totalSize - amtDone);

         sha.update(buf, 0, amt);

         nextBuf++;
         amtDone += amt;
      }

      /*
       * Now perform a second pass of the algorithm, where we compute
       * a digest value that covers the password (again) and the output
       * digest value from the first pass.
       */
      result = sha.digest();

      System.arraycopy(result, 0, destination, 0, Math.min(result.length, destination.length));
      return destination;
   }

   /**
   
   Method to encode the parameters for the remote array authentication.
   
   NOTES/ASSUMPTIONS: <BR>
   None.    <P>
   
   @param identifier          the Storage Array Identifier of the remote array.
   @param configGeneration    the config generation number for the remote array
   @param label               label of the associated entity
   @param passWordStr         the password for the remote array.
   @return
   @throws RPCError
    */
   public static byte[] encodeRemoteAuthenticator(ApiDefinition apiDef, SAIdentifier identifier,
         long configGeneration, String label, String passWordStr)
         throws RPCError
   {
      RemoteArrayAuthentication payload = new RemoteArrayAuthentication();
      byte destination[] = new byte[DIGEST_LENGTH];
      MessageDigest sha;
      XdrEncoder encoder = new XdrEncoder();
      XdrMessage  msg;
      byte password[] = getBytes(passWordStr);
      int totalSize = 0;
      int amtDone = 0;
      int nextBuf = 0;
      int amt;
      byte[] result;
      byte[] buf;

      TypeSpec authTypeSpec = apiDef.getTypeSpec(RemoteArrayAuthentication.class);
      
      // Build remote authenticator
      payload.setArrayId(identifier);
      payload.setCfgGen(configGeneration);

      String labelStrBuffer = ByteUtils.getHexString(getBytes(label));
      payload.setOperationIdentifier(labelStrBuffer);

      try
      {
         sha = MessageDigest.getInstance("SHA");
      } 
      catch (Exception ex)
      {
         throw new RPCError(ErrorType.AUTH_ALGORITHM);
      }

      sha.update(password);

      try
      {
    	  encoder.encode(payload, authTypeSpec);
      } 
      catch (EncodeException e)
      {
    	  e.printStackTrace();
         throw new RPCError(ErrorType.AUTH_ALGORITHM);
      }

      msg = encoder.getMessage();
      totalSize = msg.getMessageSize();

      /*
       * Iteratively update the digest output value by feeding the digest engine
       * with each buffer that comprises the message.
       */
      while (amtDone < totalSize)
      {
         buf = msg.getBuffer(nextBuf);
         amt = Math.min(buf.length, totalSize - amtDone);

         sha.update(buf, 0, amt);

         nextBuf++;
         amtDone += amt;
      }

      /*
       * Now perform a second pass of the algorithm, where we compute a digest
       * value that covers the password (again) and the output digest value from
       * the first pass.
       */
      result = sha.digest();

      System.arraycopy(result, 0, destination, 0, Math.min(result.length,
            destination.length));
      return destination;
   }
   
   /**
    * DESCRIPTION: <BR>
    * Convert the specified UNICODE-compliant Java string into a
    * raw byte array.
    * <BR>
    * NOTES/ASSUMPTIONS: <BR>
    * Characters are stored in the array with MSB first, followed by LSB.
    * <BR><BR>
    * @param str String to be converted to a raw byte array.
    * @return The raw byte array that corresponds to the string.
    */
   private static byte[] getBytes(String str)
   {
      int len = 2 * str.length();
      byte[] result = new byte[len];
      int i;
      int value;

      // fill in the result array
      for (i = 0; i < len / 2; i++)
      {
         value = (int) str.charAt(i);

         result[(2 * i) + 0] = (byte) (value >> 8);
         result[(2 * i) + 1] = (byte) (value >> 0);
      }

      return result;
   }
}
