/******************************************************************************
 * Copyright 2015 NetApp, Inc. All rights reserved.                           *
 *                                                                            *
 * This file is confidential and a trade secret of NetApp, Inc.               *
 * The receipt of or possession of this file does not convey any rights to    *
 * reproduce or disclose its contents or to manufacture, use, or sell         *
 * anything it may describe, in whole, or in part, without the specific       *
 * written consent of NetApp, Inc.                                            *
 ******************************************************************************/

package devmgr.api.symbolutils;

import devmgr.api.rpc.RPCError;
import devmgr.api.runtime.XdrMessage;

/**
 * Base class for generating a SYMbol Message Authentication Code.
 * <P>
 * REFERENCE: <BR>
 * http://ictwiki.eng.netapp.com/index.php/SYMbol_security_enhancement
 */
public abstract class SymbolMac {
    /**
     * Gets the size of the MAC output.
     *
     * @return size of the MAC output in bytes
     */
    public abstract int getOutputSize();


    /**
     * Sets the key to be used in the MAC. Must be called before any update or complete methods are called.
     *
     * @param key key for the MAC, generally a storage array password encoded in UCS-2
     * @throws RPCError if the key is not usable
     */
    public abstract void setKey(byte[] key) throws RPCError;


    /**
     * Adds data upon which the MAC is to be calculated. setKey method must be called before this method.
     *
     * @param input data to add to the MAC
     */
    public void update(byte[] input) {
        update(input, 0, input.length);
    }


    /**
     * Adds data upon which the MAC is to be calculated. setKey method must be called before this method.
     *
     * @param input data to add to the MAC
     * @param offset 0 relative offset into the byte array from which to start
     * @param length amount of data from input to add
     */
    public abstract void update(byte[] input, int offset, int length);


    /**
     * Adds data upon which the MAC is to be calculated. setKey method must be called before this method.
     *
     * @param message XDR message IOVector, all the data in the buffer will be added to the MAC
     */
    public void update(XdrMessage message) {
        int totalBytes = message.getMessageSize();
        int bytesCompleted = 0;
        int nextBuffer = 0;

        while (bytesCompleted < totalBytes) {
            byte[] buffer = message.getBuffer(nextBuffer);
            int currentByteCount = Math.min(buffer.length, totalBytes - bytesCompleted);

            update(buffer, 0, currentByteCount);

            nextBuffer++;
            bytesCompleted += currentByteCount;
        }
    }


    /**
     * Finishes the MAC calculations and generates the final MAC digest. setKey method must be called before this
     * method.
     *
     * @return MAC output
     * @throws RPCError if an error occurs completing the calculations
     */
    public byte[] digest() throws RPCError {
        byte[] output = new byte[getOutputSize()];
        digest(output, 0);
        return output;
    }


    /**
     * Finishes the MAC calculations and generates the final MAC digest. setKey method must be called before this
     * method.
     *
     * @param output byte array to put the generated MAC output
     * @param offset 0 relative offset into output to start putting the generated MAC output
     * @throws RPCError if an error occurs completing the calculations
     */
    public abstract void digest(byte[] output, int offset) throws RPCError;
}
